#!/usr/bin/env python
# encoding: utf-8

"""
funcutils provided some useful functions.
"""

import re
import operator


def template_render(template, model):
    """
    A simple template render.
    Example:
    >>> model = {
    ...     "name": "lyc",
    ...     "foo": {
    ...       "bar": "hello"
    ...     }
    ...   }
    >>> template = "{{foo.bar}} {{ name }}"
    >>> template_render(template, model)

    :param template: template string
    :param model: dict model
    """
    parts = []
    for part in re.split(r"{{([\.\w]+?)}}", template):
        parts.append(str(
            part if len(parts) & 1 == 0 else
            reduce(lambda m, k: m[k], part.split("."), model)))
    return "".join(parts)


def encode(s, encoding="utf-8", errors="strict"):
    """
    Auto encode unicode string to bytes.
    :param s: string
    :param encoding: bytes encoding(default:utf-8)
    :param errors: different error handling scheme
    :return: bytes
    """
    if isinstance(s, bytes):
        return s

    return s.encode(encoding, errors)


def decode(s, encoding="utf-8", errors="strict"):
    """
    Auto decode bytes to unicode string.
    :param s: string
    :param encoding: bytes encoding(default:utf-8)
    :param errors: different error handling scheme
    :return: unicode
    """
    if isinstance(s, unicode):
        return s

    return s.decode(encoding, errors)


def bytescat(s1, s2, encoding="utf-8"):
    """
    Auto encode and return s1 + s2
    :param s1: string
    :param s2: string
    :param encoding: bytes encoding(default:utf-8)
    :return: bytes
    """
    return encode(s1, encoding) + encode(s2, encoding)


def strcat(s1, s2, encoding1="utf-8", encoding2=None):
    """
    Auto decode and return s1 + s2
    :param s1: string
    :param s2: string
    :param encoding1: bytes encoding of s1(default:utf-8)
    :param encoding2: bytes encoding of s2(default:same as encoding1)
    :return: bytes
    """
    encoding2 = encoding2 or encoding1
    return decode(s1, encoding1) + decode(s2, encoding2)


def str_remove_left(s, pattern):
    """
    Remove left part which equal to pattern of a string.

    :param s: string
    :param pattern: string pattern
    """
    if s.startswith(pattern):
        return s[len(pattern):]
    return s


def str_remove_right(s, pattern):
    """
    Remove right part which equal to pattern of a string.

    :param s: string
    :param pattern: string pattern
    """
    if s.endswith(pattern):
        return s[:-len(pattern)]
    return s



#!/usr/bin/env python
# encoding: utf-8


def get_real_bases(bases):
    """
    Get real bases for types argument.
    >>> get_real_bases(None) #=> (object,)
    >>> get_real_bases(TypeA) #=> (TypeA,)
    >>> get_real_bases([TypeA, TypeB]) #=> (TypeA, TypeB)

    :param bases: type or type sequence
    """
    if bases is None:
        return (object,)
    if isinstance(bases, type):
        bases = (bases,)
    return tuple(bases)


def subtype(name, bases=None, attrs=None):
    """
    A easier way to create a type inherited from bases(default:object)
    with specified attrs.

    :param name: name of new type
    :param bases: bases class of new type
    :param attrs: class attributes of new type
    """
    return type(name, get_real_bases(bases), dict(attrs or {}))


class TypeFactory(object):
    """
    Create your type from this factory.
    >>> types_factory.NewType()
    equals:
    >>> subtype("NewType")
    """
    def __getattr__(self, name):
        return lambda bases=None, attrs=None: subtype(
            name, bases, attrs,
        )

types_factory = TypeFactory()


def subexception(name, bases=None, attrs=None):
    """
    A easier way to create an Exception

    :param name: name of new exception
    :param bases: bases class of new exception
    :param attrs: class attributes of new exception
    """
    return subtype(name, bases or [Exception], attrs)


class ExceptionFactory(object):
    """
    Create your type by this factory.
    >>> exceptions_factory.NewError()
    equals:
    >>> subexception("NewError")
    """
    def __getattr__(self, name):
        return lambda bases=None, attrs=None: subexception(
            name, bases, attrs,
        )

exceptions_factory = ExceptionFactory()


class SimpleExceptions(object):
    """
    Create and cached a simple exception.
    """
    def __getattr__(self, name):
        exception = subexception(name)
        setattr(self, name, exception)
        return exception



#!/usr/bin/env python
# encoding: utf-8

import contextlib
import logging.config
import logging


def quick_config(log_file="application.log"):
    """
    a quick config for logging

    :param log_file: the file for file handler
    """
    config = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'verbose': {
                'format': '%(asctime)s %(levelname)s %(name)s %(message)s',
                'datefmt': '[%d/%b/%Y %H:%M:%S]',
            },
            'simple': {
                'format': '%(asctime)s %(levelname)s %(name)s:%(lineno)s %(message)s',
                'datefmt': '[%d/%b/%Y %H:%M:%S]',
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
            'file': {
                'level': 'INFO',
                'class': 'logging.FileHandler',
                'filename': log_file,
                'formatter': 'verbose'
            },
        },
        'loggers': {
            '': {
                'handlers': ['console', 'file'],
                'level': 'DEBUG',
            },
        },
    }
    logging.config.dictConfig(config)


@contextlib.contextmanager
def log_disable(level=logging.CRITICAL):
    """
    A context manager to disable all logging calls of
    severity 'level' and below in block
    """
    level, logging.root.manager.disable = logging.root.manager.disable, level
    try:
        yield
    finally:
        level, logging.root.manager.disable = logging.root.manager.disable, level


def log_with_label(log_method, label):
    """
    >>> logger_waring = log_with_label(logger.waring, "waring: %s")
    >>> try:
    ...     may_raise_exception()
    ... except Exception as err:
    ...     logger_waring(err)

    :param log_method: method of a logger to pack
    :param label: label pass to log_method as the first argument
    :return: logger method packed method
    """
    import functools

    @functools.wraps(log_method)
    def logger_method(*args, **kwargs):
        return log_method(label, *args, **kwargs)

    return logger_method
